function im4d_f = tfilter_nii(fname, outputdir, Fs, varargin)
% function im4d_f = tfilter_nii(fname, outputdir, Fs, ... )
% 
% im4d_filtered = tfilter_nii(fname, outputdir_base, Fs, ...);
% 
% create and apply a IIR butterworth filter to nii or img/hdr
% files are read with load_nii. 
% output images are written in float64 spm_type(64)
%
% MANDATORY ARG
% fname = /pasth/to/somting/swua*.img
%     or
% fname = an array of cell containinf the image filename (including
% path) see doc cell or doc cellstr.
% 
% outputdir = /pasth/to/somting/ 
%            (in this example output files will be fswua*.img)
%            ouput images will be in double 64 bits
% 
% Fs = 0.4;  % Sampling Frequency in Hz (1/TR)
% 
% OPTIONAL ARG 'param', value
% ftype = 'bandpass' (defaults) % 
%          can be 'high', 'stop', 'low','bandpass'
% N   = 10;    (default)% filter Order (10 14 20) N/2 is actually used 
% Fc1 = 0.01;  (default)% First Cutoff Frequency in Hz
% Fc2 = 0.1;   (default)% Second Cutoff Frequency in Hz (used only for
%                         bandpass and stopband)
% prefix = 'f' (default)% output images prefix
% ic_function = @mean (default)% func handle applied before filtering to avoid
%               problems related to initial conditions
%               can be @median, f if f = @(x)(x(1)), etc... 
% ic_back = false (default) %after filtering add back the removed 
%           before filtering
% hdr_msg = 'tfiltered ' (default) % add a message in the description field
%                                  of the output images
% verbose = false (default) % display some output test 
% display = false (default) % display the filter and zerophase response
% 
% WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING
% WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING
%
% Output Images ar written assuming that no header interpretation is needed
% So swa images no problem, but be carefull on wa if it has been estimated
% but not written. check the input images header with spm before ("Display"). 
% This is not 100% tested so do it and let me know !!!!!!!!!!!
%
% WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING
% WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING
% 
% Example: 
% 
% fname = '/mnt/data3/hepi_prot/analysis/e2811/rest/GRE/*.nii'; 
% lf  = listfiles([fname ]);
% lf = lf(6:end); % lf is a deblanked array of cells 
%                 % (NOT an array of string)
%
% im4d_f = tfilter_nii(fname, outputdir, Fs,'N', 10, ... 
%  'prefix', 'k', 'fc1', 0.01, 'fc2', 0.1, 'ftype','bandpass', ... 
%  'ic_function', @median, 'verbose', true, 'display', true); 
% 
% or 
%
% im4d_f = tfilter_nii(lf, outputdir, Fs,'N', 14, ... 
%  'prefix', 'k', 'fc1', 0.009, 'fc2', 0.12, 'ftype','bandpass', ... 
%  'ic_function', @median, 'verbose', true, 'display', true); 
% 
% im4d_f = tfilter_nii(lf, outputdir, Fs, ... 
%  'ftype','high','fc1', 0.008 ); 
%
% Brice Fernandez - GE Healthcare - whenever 2011
%  updated Dec 2013 
%  updated Dec 2013 to use spm function for reading/writing images

p = inputParser; 
p.FunctionName = 'tfilter_nii'; 

p.addRequired('fname', @(x)(ischar(x) || iscell(x)) ); 
p.addRequired('outputdir', @isdir); 
p.addRequired('Fs', @(x)(isnumeric(x) && isscalar(x)) ); 

p.addParamValue('N', 10 , @(x)(isnumeric(x) && isscalar(x)) ); 
p.addParamValue('prefix', 'f', @isstr); 
p.addParamValue('fc1', 0.01, @(x)(isnumeric(x) && isscalar(x))); 
p.addParamValue('fc2', 0.1, @(x)(isnumeric(x) && isscalar(x))); 
p.addParamValue('ftype', 'bandpass',  ... 
  @(x)any(strcmpi(x,{'high', 'stop', 'low','bandpass'})) ); 
p.addParamValue('verbose', false ,@islogical ); 
p.addParamValue('ic_back', false, @islogical); 
p.addParamValue('display', false ,@islogical ); 
p.addParamValue('ic_function', @mean, @(x)(isa(x, 'function_handle')) ); 
p.addParamValue('hdr_msg', 'tfiltered ' ,@istring ); 

p.parse(fname, outputdir, Fs, varargin{:}); 

verbose = p.Results.verbose; 
display = p.Results.display; 

%verbose = true; 
if( verbose == true) 
  fprintf('(I) function %s input arg\n', p.FunctionName);
  disp(p.Results);  
end
% 
ftype = p.Results.ftype; 
N = p.Results.N; 

% Creating the filter 
if( strcmpi(ftype, 'high') || strcmpi(ftype, 'low'))
  % highpass and lowpass
  W = p.Results.fc1/(p.Results.Fs/2);   
else
  % stop of bandpass
  W = [p.Results.fc1 p.Results.fc2]./(p.Results.Fs/2); 
end

[zz,pp,kk] = butter(N/2, W, ftype , 'z');

[sos_var,gg] = zp2sos(zz, pp, kk);
Hd = dfilt.df2sos(sos_var, gg);
set(Hd, 'PersistentMemory', true); 

if( Hd.isstable == false)
  error(generatemsgid([p.FunctionName ':NotStable']), 'Butterworth filter not stable');
end

if( display == true)
  h = fvtool(Hd);
  set(h, 'Analysis', 'freq');
  set(h, 'WindowStyle', 'normal');

  set(h, 'DockControls', 'off');
  
  zerophase(Hd);
  
  drawnow; 
  %disp(['(I) Press enter to continue...']); 
  %pause; 
  %close(h);   
end

if( iscell(fname) == true)
  % here we consider that is already a list of cells
  % containing deblank ordered file name
  lf = deblank(fname); 
else
  lf = listfiles(deblank(fname));
end

lf = deblank(lf); 

if( iscell(lf) == false) 
  error( generatemsgid([ p.FunctionName ':fname']), 'fname/lf is not a cell nor a single string'); 
end

if(verbose == true)
  disp(['(I) inputs files are:']);
  disp(lf(:));
end

nfiles = length(lf); 

if( outputdir(end) ~= '/')
    outputdir = [outputdir '/']; 
end

% load images 
% originally I used load_nii I switch to spm functions
% nii1 = load_nii(lf{1}); 

% read nii hdr
nii1_hdr = spm_vol(lf{1});

% to take into account the n 3D or one 4D file
Nt = max(nfiles, length(nii1_hdr)); 

im4d = zeros([nii1_hdr(1).dim Nt]); 

if ( nfiles > 1 )
  % consiser multiple 3D files 
  nii_hdr = cell(nfiles, 1); 
  
  % multiplt nii or img/hdr files
  for k = 1 : nfiles
    %tmpnii = load_nii(lf{k}); 
    % read all nii hdr
    nii_hdr{k} = spm_vol(lf{k});
    % read images 
    nii_img = spm_read_vols(nii_hdr{k});

    im4d(:,:,:,k) = double(nii_img);  
  end
else
  if ( nfiles == 1 )
    % one 4D nii
    % read all nii hdr
    nii_hdr{1} = spm_vol(lf{1});
    % read images 
    nii_img = spm_read_vols(nii_hdr{1});
    
    im4d = double(nii_img); 
  else
    % too many dimension 
    error( generatemsgid('temporal_filters_nii:xDniiFile'), ...
      'too many dim in a nii');
  end
end



% apply the filter 
% 
im4d_f = zeros(size(im4d), 'double'); 

% get the function handler for initial condition pbm
f = p.Results.ic_function; 
ic_back = p.Results.ic_back; 

% just to display progress without using too much CPU
fprintf('Filtering...\n');
fprintf('   ***\n\n');

[nx,ny,nz,nt] = size(im4d); 
nfiltering = nx*ny*nz; 
if( verbose == true)
  mod_factor=10;
else 
  mod_factor=25; 
end
  
idx_display=0; 
% end display part 
%if(0) % debug
for x = 1 : nx, 
    for y = 1 : ny, 
        for z = 1 : nz, 
                    
          %Hd = dfilt.df2sos(sos_var, gg);
          %set(Hd, 'PersistentMemory', true);
          Hd.reset; 
          xl = squeeze(im4d(x,y,z,:));
          
          x1=xl-f(xl); 

          y1 = Hd.filter(x1);
          y1 = flipud(y1);
          y1 = Hd.filter(y1);
          y1 = flipud(y1);
          
          if( ic_back == true)
            y1 = y1 + f(xl); 
          end
          
          im4d_f(x,y,z,:) = y1;
         
          % display progress
          idx_display = idx_display + 1;
          if( mod(idx_display, mod_factor) == 0)
            progress = idx_display/nfiltering;
            fprintf('\b\b\b\b\b%03d %%',round(progress*100));
            %fprintf('%03d %%\n',progress*100);
            %waitbar(progress,h,['Loading images... ' num2str(progress, '%d')] )
          end
          
        end
    end
end

fprintf('   ***\n');

%end % debug 
%im4d_f=im4d;  % debug

% write those images
% always consider the same header
prefix = p.Results.prefix; 
hdr_msg = p.Results.hdr_msg; 

if ( nfiles > 1 )
  
  for k = 1 : nfiles,
    % get header information 
    hdrInfo = nii_hdr{k}; 
    % k
    % contruct output file name
    idx=findstr(hdrInfo.fname, '/');
    outfname = [outputdir prefix hdrInfo.fname(idx(length(idx))+1:end) ];
    
    % modify input hdr to reuse the information   
    hdrInfo.fname = outfname;
    hdrInfo.private.dat.fname = outfname;
    hdrInfo.dt(1) = 64; % spm_type(64) is float64
    hdrInfo = rmfield(hdrInfo, 'pinfo'); %remove field (see spm_write_vol)
    hdrInfo.descrip = [hdr_msg hdrInfo.descrip];
    
    spm_write_vol(hdrInfo,double(squeeze(im4d_f(:,:,:,k))));
    
  end
  
elseif ( nfiles == 1 )
  k=1;
  % get header information
  hdrInfo = nii_hdr{k};
  
  % contruct output file name
  idx=findstr(hdrInfo(1).fname, '/');
  outfname = [outputdir prefix hdrInfo(1).fname(idx(length(idx))+1:end) ];
  
  for k = 1 : length(hdrInfo),
    % modify input hdr to reuse the information
    hdrInfo(k).fname = outfname;
    hdrInfo(k).private.dat.fname = outfname;
    hdrInfo(k).dt(1) = 64; % spm_type(64) is float64
    hdrInfo(k).descrip = [hdr_msg hdrInfo(k).descrip];
  end
  hdrInfo = rmfield(hdrInfo, 'pinfo'); %remove field (see spm_write_vol)
  disp([ 'writing']);
  for k = 1 : length(hdrInfo),
    spm_write_vol(hdrInfo(k),im4d_f(:,:,:,k));
  end
else
  error( generatemsgid([pFunctionName ':xDniiOutputFile']), ...
      'too many dim for a nii this should not happen');
end
    
    
% Original code using load_nii/save_nii
%nii = make_nii(im4d_f, voxel_size, origin, datatype, description);
%niiout.hdr = nii1.hdr;
% convert to floating point double
% convert to floating point double
%niiout.hdr.dime.datatype=64;
%niiout.hdr.dime.bitpix = 64;
%niiout.img = im4d_f(:,:,:,k);
%save_nii(niiout, outfname);

    
    
    
    
    
    
    